#ifndef GUARD_99E25E7F8183EF0D8D27390EBC173C04
#define GUARD_99E25E7F8183EF0D8D27390EBC173C04

#include "types.h"

typedef u32 sw_color32;

sw_color32 sw_color32_from_rgb(u8 r, u8 g, u8 b);
sw_color32 sw_color32_from_rgba(u8 r, u8 g, u8 b, u8 a);

u8 sw_color32_get_r(sw_color32 col);
u8 sw_color32_get_g(sw_color32 col);
u8 sw_color32_get_b(sw_color32 col);
u8 sw_color32_get_a(sw_color32 col);

f32 sw_color32_get_rf(sw_color32 col);
f32 sw_color32_get_gf(sw_color32 col);
f32 sw_color32_get_bf(sw_color32 col);
f32 sw_color32_get_af(sw_color32 col);

#endif /* GUARD_99E25E7F8183EF0D8D27390EBC173C04 */
